#!/usr/bin/env -S nim c -d:release
import sequtils, strutils, strformat
import std/[times, monotimes]
type Timer = object
    id: string
    start: MonoTime
proc `=destroy`(t: var Timer) =
    echo t.id & ": " & (inMilliseconds(getMonoTime()-t.start).float/1000.0).formatFloat(ffDecimal, 3)
template time(timer_id: string) =
  var Timer = Timer(id:timer_id, start:getMonoTime())
import os, osproc
proc cmd*(c: string): string {.discardable.} = execProcess(c)
import hashes
import tables
import math

let
  avr = getAppDir() / "avr.sh "
  btorsim = getAppDir() / "avr/deps/btor2tools/build/bin/btorsim "
  boolector = getAppDir() / "avr/deps/boolector/build/bin/boolector "
type Op = enum
  Sort = "sort",

  One = "one", Ones = "ones", Zero = "zero", Input = "input", State = "state",
  Bad = "bad", Output = "output", Constraint = "constraint", Fair = "fair", Justice = "justice"

  Const = "const", Constd = "constd", Consth = "consth", Dec = "dec", Inc = "inc",
  Neg = "neg", Not = "not", Redand = "redand", Redor = "redor", Redxor = "redxor",
  Sext = "sext", Uext = "uext", Slice = "slice"

  Add = "add", And = "and", Concat = "concat", Eq = "eq", Implies = "implies",
  Iff = "iff", Mul = "mul", Nand = "nand", Neq = "neq", Nor = "nor",
  Or = "or", Read = "read", Rol = "rol", Ror = "ror", Saddo = "saddo",
  Sgt = "sgt", Sgte = "sgte", Sdiv = "sdiv", Sdivo = "sdivo", Sll = "sll",
  Slt = "slt", Slte = "slte", Smod = "smod", Smulo = "smulo", Ssubo = "ssubo",
  Sra = "sra", Srl = "srl", Srem = "srem", Sub = "sub", Uaddo = "uaddo",
  Udiv = "udiv", Ugt = "ugt", Ugte = "ugte", Ult = "ult", Ulte = "ulte",
  Umulo = "umulo", Urem = "urem", Usubo = "usubo", Xnor = "xnor", Xor = "xor",
  Init = "init", Next = "next"

  Ite = "ite", Write = "write"

  Nop
const arity = [
  0,
  1,1,1,1,1,1,1,1,1,1,1,1,1,
  2,2,2,2,2,2,2,2,2,2,
  3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,3,
  4,4
  ]
type Btor = seq[seq[int]]
var
  offset = 0
  encoding: Btor
  Bool: int
  constTable: seq[string]

var params = commandLineParams()
if params.len < 2: quit &"Usage: {paramStr 0} <path_to_model> k"
let
  pathToModel = params[0]
  name = pathToModel.extractFilename.changeFileExt ""
  k = parseInt params[1]
echo "name: " & name
echo "k: " & $k

proc iToString(i: seq[int]): string =
  result = $i[0] & " "
  if i[1].Op == Sort: result &= "sort bitvec"
  else: result &= $i[1].Op
  if i[1].Op == Const:
    result &= " " & $i[2] & " " & constTable[i[3]]
  else:
    for e in i[2..^1]: result &= " " & $e

var tempFiles: seq[string]
proc tempFile(ext: string):string =
  result = (getTempDir()/name)&"_"&ext.addFileExt "btor2"
  tempFiles.add result
proc writeToFile(encoding: Btor, ext: string) =
  echo &"s_{ext}: {len encoding}"
  ext.tempFile.writeFile encoding.map(iToString).join("\n") & "\n"
  let r = cmd btorsim&ext.tempFile
  if "error" in r:
    quit &"error simulating {ext}\n{r}"
proc assertUnsat(encoding: Btor, ext: string) =
  writeToFile(encoding, ext)
  time "t_"&ext
  let r = cmd boolector&tempFile(ext)
  if "unsat" notin r:
       quit &"error check failed {ext}\n{r}"

proc l(i: varargs[int, int]): int {.discardable.} =
  result = (if encoding.len == 0: 1
            else: encoding[^1][0]+1)
  var r = @[result]
  for e in i: r.add e
  encoding.add r
proc last: int = (if encoding.len == 0: 0 else: encoding[^1][0])
proc shift(i: int): int = i+offset
proc shift(i: seq[int]): int {.discardable.} =
  result = shift i[0]
  encoding.add @[result, i[1]] &
    i[2..<2+arity[i[1]]].map(shift) &
    i[2+arity[i[1]]..^1]

proc reduce(c: seq[int], op: Op): int =
  var c = c
  while c.len > 1:
    for i in countup(0, c.high - 1, 2): c[i div 2] = l(op, Bool, c[i], c[i+1])
    if c.len mod 2 == 1: c[(c.len+1) div 2 - 1] = c[^1]
    c.setLen (c.len+1) div 2
  return c[^1]
proc conj(c: seq[int]): int = c.reduce And
proc disj(c: seq[int]): int = c.reduce Or

template encode(id: string) =
  time id
  encoding = @[]
  offset = 0
  Bool = -1

proc readModel(pathToModel: string): Btor =
  encode "r_model"
  let file = pathToModel.readfile.strip.splitlines
  for l in file:
    if l[0] == ';': continue
    let l = split l
    var o = Nop
    try: o = parseEnum[Op] l[1]
    except ValueError: quit &"Operator {l[1]} not covered"
    if   o == Sort: encoding.add @[l[0].parseInt, Sort.int, l[3].parseInt]
    elif o == Const:
      constTable.add l[3]
      encoding.add @[l[0].parseInt, Const.int, l[2].parseInt, constTable.high]
    else:
      var r = @[l[0].parseInt, o.int]
      for e in l[2..^1]:
        if not isDigit e[0]: break
        r.add parseInt e
      encoding.add r
  return encoding

proc k_witness(model: Btor, k: int): Btor =
  encode "g_witness"
  var
    inputs = newSeq[seq[tuple[ind, sort: int]]] k
    states = newSeq[seq[tuple[ind, sort, next, init: int]]] k
    sMap: Table[int, int]
    bad = newSeq[int] k
  for j in countdown(k-1, 0):
    offset = last()
    for i in model:
      if Bool == -1 and i[1].Op == Sort and i[2] == 1:
        Bool = shift i[0]
      case i[1].Op:
        of Input:
          inputs[j].add (
            shift @[i[0], (
              if j==k-1: Input.int else: State.int),
                    i[2]], shift i[2])
        of State:
          states[j].add (shift i, shift i[2], -1, -1)
          if j==k-1: sMap[i[0]] = high states[j]
        of Next:
          states[j][sMap[i[3]]].next = shift i[4]
          if j==k-1: shift i
        of Init:
          states[j][sMap[i[3]]].init = shift i[4]
          if j==k-1: shift i
        of Bad:
          bad[j] = shift i[2]
        else: shift i
    if j==k-1: continue
    for (c,n) in zip(inputs[j], inputs[j+1]):
      l Next, c.sort, c.ind, n.ind
    for (c,n) in zip(states[j], states[j+1]):
      l Next, c.sort, c.ind, n.ind

  let typeB = l(Sort, k.float.log2.ceil.int)
  var index = (-1).repeat k
  for j in countdown(k-1, 0):
    index[j] = l(Constd, typeB, j)
  let B = l(State, typeB)
  l Init, typeB, B, index[k-1]
  let sub = l(Sub, typeB, B, index[1])
  l Eq, Bool, B, index[0]
  l Ite, typeB, last(), index[0], sub
  l Next, typeB, B, last()
  var uninitialized = (-1).repeat k
  for j in countdown(k-1, 0):
    uninitialized[j] = l(Ult, Bool, index[j], B)

  var prop: seq[int]
  prop.add l(Ugte, Bool, B, index[0])
  prop.add l(Ulte, Bool, B, index[k-1])

  for j in countdown(k-2, 0):
    for (c,n) in zip(states[j], states[j+1]):
      if c.next == -1: continue
      l Eq, Bool, c.next, n.ind
      prop.add l(Or, Bool, last(), uninitialized[j])

  for j in countdown(k-1, 0):
    l Not, Bool, bad[j]
    prop.add l(Or, Bool, uninitialized[j], last())

  for j in countdown(k-1, 1):
    var reset: seq[int]
    for c in states[j]:
      if c.init == -1: continue
      reset.add l(Eq, Bool, c.ind, c.init)
    # TODO empty reset
    let isReset = conj(reset)
    l Neq, Bool, index[j], B
    prop.add l(Or, Bool, last(), isReset)

  l Not, Bool, conj prop
  l Bad, last()
  return encoding

proc consecution(witness: Btor): Btor =
  encode "g_consec"
  var prop: seq[int]
  for isCopy in [false, true]:
    offset = last()
    for i in witness:
      if Bool == -1 and i[1].Op == Sort and i[2] == 1:
        Bool = shift i[0]
      case i[1].Op:
        of State:
          shift @[i[0], Input.int, i[2]]
        of Next:
          if isCopy: prop.add l(Eq, Bool, shift i[3], i[4])
        of Init: discard
        of Bad: prop.add (
          if isCopy: shift i[2]
          else: l(Not, Bool, shift i[2]))
        else: shift i
  l Constraint, conj prop
  return encoding

proc initialization(witness: Btor): Btor =
  encode "g_init"
  var prop: seq[int]
  Bool = l(Sort, 1)
  offset = last()
  for i in witness:
    if Bool == -1 and i[1].Op == Sort and i[2] == 1:
      Bool = shift i[0]
    case i[1].Op:
      of State:
        shift @[i[0], Input.int, i[2]]
      of Next: discard
      of Init:
        prop.add l(Eq, Bool, shift i[3], shift i[4])
      of Bad: prop.add shift i[2]
      else: shift i
  l Constraint, conj prop
  return encoding

proc consistency(witness: Btor): Btor =
  encode "g_consist"
  var
    inputs = [newSeq[int](), newSeq[int]()]
    bad = [-1, -1]
  for j in 0..1:
    offset = last()
    for i in witness:
      if Bool == -1 and i[1].Op == Sort and i[2] == 1:
        Bool = shift i[0]
      case i[1].Op:
        of Input:
          inputs[j].add shift i
        of State:
          inputs[j].add shift @[i[0], Input.int, i[2]]
        of Next: discard
        of Init: discard
        of Bad: bad[j] = shift i[2]
        else: shift i
  var prop: seq[int]
  for (a, b) in zip(inputs[0], inputs[1]):
    prop.add l(Eq, Bool, a, b)
  prop.add bad[0]
  prop.add l(Not, Bool, bad[1])
  l Constraint, conj prop
  return encoding

proc transition(model, witness: Btor): Btor =
  encode "g_trans"
  var
    inputs: array[2, seq[tuple[ind, next: int]]]
    iMap: Table[int, int]
  for j, machine in [model, witness]:
    offset = last()
    for i in machine:
      if Bool == -1 and i[1].Op == Sort and i[2] == 1:
        Bool = shift i[0]
      case i[1].Op:
        of Input:
          inputs[j].add (shift i, -1)
          if j==0: iMap[i[0]] = high inputs
        of State:
          inputs[j].add (shift @[i[0], Input.int, i[2]], -1)
          if j==0: iMap[i[0]] = high inputs
        of Next:
          if i[3] in iMap:
            inputs[j][iMap[i[3]]].next = shift i[4]
        of Init: discard
        of Bad: discard
        else: shift i
  var prop: seq[int]
  for (a, b) in zip(inputs[0], inputs[1]):
    if a.next == -1: continue
    prop.add l(Neq, Bool, a.next, b.next)
  # TODO: what if no state has a transition?
  prop = @[disj(prop)]
  for (a, b) in zip(inputs[0], inputs[1]):
    prop.add l(Eq, Bool, a.ind, b.ind)
  l Constraint, conj prop
  return encoding

proc property(model, witness: Btor): Btor =
  encode "g_prop"
  var
    inputs = [newSeq[int](), newSeq[int]()]
    bad = [-1, -1]
  for j, machine in [model, witness]:
    offset = last()
    for i in machine:
      if Bool == -1 and i[1].Op == Sort and i[2] == 1:
        Bool = shift i[0]
      case i[1].Op:
        of Input:
          inputs[j].add shift i
        of State:
          inputs[j].add shift @[i[0], Input.int, i[2]]
        of Next: discard
        of Init: discard
        of Bad: bad[j] = shift i[2]
        else: shift i
  var prop: seq[int]
  for (a, b) in zip(inputs[0], inputs[1]):
    prop.add l(Eq, Bool, a, b)
  prop.add bad[0]
  prop.add l(Not, Bool, bad[1])
  l Constraint, conj prop
  return encoding

proc reset(model, witness: Btor): Btor =
  encode "g_reset"
  var
    inputs: array[2, seq[tuple[ind, init: int]]]
    iMap: Table[int, int]
  for j, machine in [model, witness]:
    offset = last()
    for i in machine:
      if Bool == -1 and i[1].Op == Sort and i[2] == 1:
        Bool = shift i[0]
      case i[1].Op:
        of Input:
          inputs[j].add (shift i, -1)
          if j==0: iMap[i[0]] = high inputs
        of State:
          inputs[j].add (shift @[i[0], Input.int, i[2]], -1)
          if j==0: iMap[i[0]] = high inputs
        of Next: discard
        of Init:
          if i[3] in iMap:
            inputs[j][iMap[i[3]]].init = shift i[4]
        of Bad: discard
        else: shift i
  var prop: seq[int]
  for (a, b) in zip(inputs[0], inputs[1]):
    if a.init == -1: continue
    prop.add l(Neq, Bool, a.init, b.init)
  # TODO: what if no state has an init?
  prop = @[disj(prop)]
  for (a, b) in zip(inputs[0], inputs[1]):
    prop.add l(Eq, Bool, a.ind, b.ind)
  l Constraint, conj prop
  return encoding

time "t_cert"
let model = readModel(pathToModel)
echo &"s_model: {len model}"
if k == 1:
  assertUnsat(consecution(model), "consec")
  assertUnsat(initialization(model), "init")
  assertUnsat(consistency(model), "consist")
else:
  let witness = k_witness(model, k)
  writeToFile(witness, "wit")
  # assert "kind: inductive at step 1" in cmd avr&tempFile "wit" #expensive
  assertUnsat(consecution(witness), "consec")
  assertUnsat(initialization(witness), "init")
  assertUnsat(consistency(witness), "consist")
  assertUnsat(transition(model, witness), "trans")
  assertUnsat(property(model, witness), "prop")
  assertUnsat(reset(model, witness), "reset")
echo "Verification Successful"

for f in tempFiles: cmd "rm " & f
